@@ -18,6 +18,7 @@ import (
 	"bytes"
 	"crypto/tls"
 	"crypto/x509"
+	"errors"
 	"fmt"
 	"net"
 	"os"
@@ -234,6 +235,103 @@ func TestServerSecureConnections(t *testing.T) {
 	}
 }
 
+func TestClientTLSConfig(t *testing.T) {
+	s, opts := RunServerWithConfig("./configs/tlsverify.conf")
+	defer s.Shutdown()
+
+	endpoint := fmt.Sprintf("%s:%d", opts.Host, opts.Port)
+	secureURL := fmt.Sprintf("nats://%s", endpoint)
+
+	// Make sure this fails
+	nc, err := nats.Connect(secureURL, nats.Secure())
+	if err == nil {
+		nc.Close()
+		t.Fatal("Should have failed (TLS) connection without client certificate")
+	}
+	cert, err := os.ReadFile("./configs/certs/client-cert.pem")
+	if err != nil {
+		t.Fatal("Failed to read client certificate")
+	}
+	key, err := os.ReadFile("./configs/certs/client-key.pem")
+	if err != nil {
+		t.Fatal("Failed to read client key")
+	}
+	rootCAs, err := os.ReadFile("./configs/certs/ca.pem")
+	if err != nil {
+		t.Fatal("Failed to read root CAs")
+	}
+
+	certCB := func() (tls.Certificate, error) {
+		cert, err := tls.X509KeyPair(cert, key)
+		if err != nil {
+			return tls.Certificate{}, fmt.Errorf("nats: error loading client certificate: %w", err)
+		}
+		cert.Leaf, err = x509.ParseCertificate(cert.Certificate[0])
+		if err != nil {
+			return tls.Certificate{}, fmt.Errorf("nats: error parsing client certificate: %w", err)
+		}
+		return cert, nil
+	}
+
+	caCB := func() (*x509.CertPool, error) {
+		pool := x509.NewCertPool()
+		ok := pool.AppendCertsFromPEM(rootCAs)
+		if !ok {
+			return nil, fmt.Errorf("nats: failed to parse root certificate from")
+		}
+		return pool, nil
+	}
+
+	// Check parameters validity
+	_, err = nats.Connect(secureURL, nats.ClientTLSConfig(nil, nil))
+	if !errors.Is(err, nats.ErrClientCertOrRootCAsRequired) {
+		t.Fatalf("Expected error %q, got %q", nats.ErrClientCertOrRootCAsRequired, err)
+	}
+
+	certErr := &tls.CertificateVerificationError{}
+	// Should fail because of missing CA
+	_, err = nats.Connect(secureURL,
+		nats.ClientCert("./configs/certs/client-cert.pem", "./configs/certs/client-key.pem"))
+	if ok := errors.As(err, &certErr); !ok {
+		t.Fatalf("Expected error %q, got %q", nats.ErrClientCertOrRootCAsRequired, err)
+	}
+
+	// Should fail because of missing certificate
+	_, err = nats.Connect(secureURL,
+		nats.ClientTLSConfig(nil, caCB))
+	if !strings.Contains(err.Error(), "bad certificate") && !strings.Contains(err.Error(), "certificate required") {
+		t.Fatalf("Expected missing certificate error; got: %s", err)
+	}
+
+	nc, err = nats.Connect(secureURL,
+		nats.ClientTLSConfig(certCB, caCB))
+	if err != nil {
+		t.Fatalf("Failed to create (TLS) connection: %v", err)
+	}
+	defer nc.Close()
+
+	omsg := []byte("Hello!")
+	checkRecv := make(chan bool)
+
+	received := 0
+	nc.Subscribe("foo", func(m *nats.Msg) {
+		received++
+		if !bytes.Equal(m.Data, omsg) {
+			t.Fatal("Message received does not match")
+		}
+		checkRecv <- true
+	})
+	err = nc.Publish("foo", omsg)
+	if err != nil {
+		t.Fatalf("Failed to publish on secure (TLS) connection: %v", err)
+	}
+	nc.Flush()
+
+	if err := Wait(checkRecv); err != nil {
+		t.Fatal("Failed to receive message")
+	}
+}
+
 func TestClientCertificate(t *testing.T) {
 	s, opts := RunServerWithConfig("./configs/tlsverify.conf")
 	defer s.Shutdown()
@@ -996,16 +1094,21 @@ func TestCallbacksOrder(t *testing.T) {
 }
 
 func TestConnectHandler(t *testing.T) {
+	handler := func(ch chan bool) func(*nats.Conn) {
+		return func(*nats.Conn) {
+			ch <- true
+		}
+	}
 	t.Run("with RetryOnFailedConnect, connection established", func(t *testing.T) {
 		s := RunDefaultServer()
 		defer s.Shutdown()
 
 		connected := make(chan bool)
-		connHandler := func(*nats.Conn) {
-			connected <- true
-		}
+		reconnected := make(chan bool)
+
 		nc, err := nats.Connect(nats.DefaultURL,
-			nats.ConnectHandler(connHandler),
+			nats.ConnectHandler(handler(connected)),
+			nats.ReconnectHandler(handler(reconnected)),
 			nats.RetryOnFailedConnect(true))
 
 		if err != nil {
@@ -1015,24 +1118,28 @@ func TestConnectHandler(t *testing.T) {
 		if err = Wait(connected); err != nil {
 			t.Fatal("Timeout waiting for connect handler")
 		}
+		if err = WaitTime(reconnected, 100*time.Millisecond); err == nil {
+			t.Fatal("Reconnect handler should not have been invoked")
+		}
 	})
 	t.Run("with RetryOnFailedConnect, connection failed", func(t *testing.T) {
 		connected := make(chan bool)
-		connHandler := func(*nats.Conn) {
-			connected <- true
-		}
+		reconnected := make(chan bool)
+
 		nc, err := nats.Connect(nats.DefaultURL,
-			nats.ConnectHandler(connHandler),
+			nats.ConnectHandler(handler(connected)),
+			nats.ReconnectHandler(handler(reconnected)),
 			nats.RetryOnFailedConnect(true))
 
 		if err != nil {
 			t.Fatalf("Unexpected error: %v", err)
 		}
 		defer nc.Close()
-		select {
-		case <-connected:
-			t.Fatalf("ConnectedCB invoked when no connection established")
-		case <-time.After(100 * time.Millisecond):
+		if err = WaitTime(connected, 100*time.Millisecond); err == nil {
+			t.Fatal("Connected handler should not have been invoked")
+		}
+		if err = WaitTime(reconnected, 100*time.Millisecond); err == nil {
+			t.Fatal("Reconnect handler should not have been invoked")
 		}
 	})
 	t.Run("no RetryOnFailedConnect, connection established", func(t *testing.T) {
@@ -1040,11 +1147,11 @@ func TestConnectHandler(t *testing.T) {
 		defer s.Shutdown()
 
 		connected := make(chan bool)
-		connHandler := func(*nats.Conn) {
-			connected <- true
-		}
+		reconnected := make(chan bool)
 		nc, err := nats.Connect(nats.DefaultURL,
-			nats.ConnectHandler(connHandler))
+			nats.ConnectHandler(handler(connected)),
+			nats.ReconnectHandler(handler(reconnected)))
+
 		if err != nil {
 			t.Fatalf("Unexpected error: %v", err)
 		}
@@ -1052,22 +1159,94 @@ func TestConnectHandler(t *testing.T) {
 		if err = Wait(connected); err != nil {
 			t.Fatal("Timeout waiting for connect handler")
 		}
+		if err = WaitTime(reconnected, 100*time.Millisecond); err == nil {
+			t.Fatal("Reconnect handler should not have been invoked")
+		}
 	})
 	t.Run("no RetryOnFailedConnect, connection failed", func(t *testing.T) {
 		connected := make(chan bool)
-		connHandler := func(*nats.Conn) {
-			connected <- true
-		}
+		reconnected := make(chan bool)
 		_, err := nats.Connect(nats.DefaultURL,
-			nats.ConnectHandler(connHandler))
+			nats.ConnectHandler(handler(connected)),
+			nats.ReconnectHandler(handler(reconnected)))
 
 		if err == nil {
 			t.Fatalf("Expected error on connect, got nil")
 		}
-		select {
-		case <-connected:
-			t.Fatalf("ConnectedCB invoked when no connection established")
-		case <-time.After(100 * time.Millisecond):
+		if err = WaitTime(connected, 100*time.Millisecond); err == nil {
+			t.Fatal("Connected handler should not have been invoked")
+		}
+		if err = WaitTime(reconnected, 100*time.Millisecond); err == nil {
+			t.Fatal("Reconnect handler should not have been invoked")
+		}
+	})
+	t.Run("with RetryOnFailedConnect, initial connection failed, reconnect successful", func(t *testing.T) {
+		connected := make(chan bool)
+		reconnected := make(chan bool)
+
+		nc, err := nats.Connect(nats.DefaultURL,
+			nats.ConnectHandler(handler(connected)),
+			nats.ReconnectHandler(handler(reconnected)),
+			nats.RetryOnFailedConnect(true),
+			nats.ReconnectWait(100*time.Millisecond))
+
+		if err != nil {
+			t.Fatalf("Expected error on connect, got nil")
+		}
+
+		defer nc.Close()
+
+		s := RunDefaultServer()
+		defer s.Shutdown()
+
+		if err != nil {
+			t.Fatalf("Expected error on connect, got nil")
+		}
+		if err = Wait(connected); err != nil {
+			t.Fatal("Timeout waiting for reconnect handler")
+		}
+		if err = WaitTime(reconnected, 100*time.Millisecond); err == nil {
+			t.Fatal("Reconnect handler should not have been invoked")
+		}
+	})
+	t.Run("with RetryOnFailedConnect, initial connection successful, server restart", func(t *testing.T) {
+		connected := make(chan bool)
+		reconnected := make(chan bool)
+
+		s := RunDefaultServer()
+		defer s.Shutdown()
+
+		nc, err := nats.Connect(nats.DefaultURL,
+			nats.ConnectHandler(handler(connected)),
+			nats.ReconnectHandler(handler(reconnected)),
+			nats.RetryOnFailedConnect(true),
+			nats.ReconnectWait(100*time.Millisecond))
+
+		if err != nil {
+			t.Fatalf("Unexpected error: %v", err)
+		}
+		defer nc.Close()
+
+		if err != nil {
+			t.Fatalf("Unexpected error: %v", err)
+		}
+		if err = Wait(connected); err != nil {
+			t.Fatal("Timeout waiting for connect handler")
+		}
+		if err = WaitTime(reconnected, 100*time.Millisecond); err == nil {
+			t.Fatal("Reconnect handler should not have been invoked")
+		}
+
+		s.Shutdown()
+
+		s = RunDefaultServer()
+		defer s.Shutdown()
+
+		if err = Wait(reconnected); err != nil {
+			t.Fatal("Timeout waiting for reconnect handler")
+		}
+		if err = WaitTime(connected, 100*time.Millisecond); err == nil {
+			t.Fatal("Connected handler should not have been invoked")
 		}
 	})
 }
@@ -1668,8 +1847,8 @@ func TestDefaultOptionsDialer(t *testing.T) {
 	s := RunDefaultServer()
 	defer s.Shutdown()
 
-	opts1 := nats.DefaultOptions
-	opts2 := nats.DefaultOptions
+	opts1 := nats.GetDefaultOptions()
+	opts2 := nats.GetDefaultOptions()
 
 	nc1, err := opts1.Connect()
 	if err != nil {
@@ -2391,7 +2570,7 @@ func TestConnectWithSimplifiedURLs(t *testing.T) {
 	opts.Username = ""
 	opts.Password = ""
 	// and restart the server
-	s = RunServerWithOptions(*opts)
+	s = RunServerWithOptions(opts)
 	defer s.Shutdown()
 
 	// Test again against a server that wants TLS and check
@@ -2435,7 +2614,7 @@ func TestGetClientID(t *testing.T) {
 	optsA.Cluster.Port = -1
 	optsA.Cluster.Name = "test"
 
-	srvA := RunServerWithOptions(optsA)
+	srvA := RunServerWithOptions(&optsA)
 	defer srvA.Shutdown()
 
 	ch := make(chan bool, 1)
@@ -2466,7 +2645,7 @@ func TestGetClientID(t *testing.T) {
 	optsB.Cluster.Name = "test"
 
 	optsB.Routes = server.RoutesFromStr(fmt.Sprintf("nats://127.0.0.1:%d", srvA.ClusterAddr().Port))
-	srvB := RunServerWithOptions(optsB)
+	srvB := RunServerWithOptions(&optsB)
 	defer srvB.Shutdown()
 
 	// Wait for the discovered callback to fire
@@ -2611,7 +2790,8 @@ func TestRetryOnFailedConnect(t *testing.T) {
 		nc.Close()
 		t.Fatal("Expected error, did not get one")
 	}
-	ch := make(chan bool, 1)
+	reconnectedCh := make(chan bool, 1)
+	connectedCh := make(chan bool, 1)
 	dch := make(chan bool, 1)
 	nc, err = nats.Connect(nats.DefaultURL,
 		nats.RetryOnFailedConnect(true),
@@ -2620,8 +2800,11 @@ func TestRetryOnFailedConnect(t *testing.T) {
 		nats.DisconnectErrHandler(func(_ *nats.Conn, _ error) {
 			dch <- true
 		}),
+		nats.ConnectHandler(func(_ *nats.Conn) {
+			connectedCh <- true
+		}),
 		nats.ReconnectHandler(func(_ *nats.Conn) {
-			ch <- true
+			reconnectedCh <- true
 		}),
 		nats.NoCallbacksAfterClientClose())
 	if err != nil {
@@ -2639,19 +2822,19 @@ func TestRetryOnFailedConnect(t *testing.T) {
 		s := RunDefaultServer()
 		defer s.Shutdown()
 
-		var action string
 		switch i {
 		case 0:
-			action = "connected"
+			select {
+			case <-connectedCh:
+			case <-time.After(2 * time.Second):
+				t.Fatal("Should have connected")
+			}
 		case 1:
-			action = "reconnected"
-		}
-
-		// Wait for the reconnect CB which in this context means that we connected ok
-		select {
-		case <-ch:
-		case <-time.After(2 * time.Second):
-			t.Fatalf("Should have %s", action)
+			select {
+			case <-reconnectedCh:
+			case <-time.After(2 * time.Second):
+				t.Fatal("Should have reconnected")
+			}
 		}
 
 		// Now make sure that the pub worked and sub worked.
@@ -2684,7 +2867,7 @@ func TestRetryOnFailedConnect(t *testing.T) {
 		nats.MaxReconnects(-1),
 		nats.ReconnectWait(15*time.Millisecond),
 		nats.ReconnectHandler(func(_ *nats.Conn) {
-			ch <- true
+			reconnectedCh <- true
 		}),
 		nats.ClosedHandler(func(_ *nats.Conn) {
 			closedCh <- true
@@ -2699,7 +2882,7 @@ func TestRetryOnFailedConnect(t *testing.T) {
 	o.Port = 4222
 	o.Username = "user"
 	o.Password = "password"
-	s := RunServerWithOptions(o)
+	s := RunServerWithOptions(&o)
 	defer s.Shutdown()
 
 	select {
@@ -2709,7 +2892,7 @@ func TestRetryOnFailedConnect(t *testing.T) {
 	}
 	// Make sure that we did not get the (re)connected CB
 	select {
-	case <-ch:
+	case <-reconnectedCh:
 		t.Fatal("(re)connected callback should not have been invoked")
 	default:
 	}
@@ -2729,17 +2912,17 @@ func TestRetryOnFailedConnectWithTLSError(t *testing.T) {
 	}
 	opts.TLSTimeout = 0.0001
 
-	s := RunServerWithOptions(opts)
+	s := RunServerWithOptions(&opts)
 	defer s.Shutdown()
 
-	ch := make(chan bool, 1)
+	connectedCh := make(chan bool, 1)
 	nc, err := nats.Connect(nats.DefaultURL,
 		nats.Secure(&tls.Config{InsecureSkipVerify: true}),
 		nats.RetryOnFailedConnect(true),
 		nats.MaxReconnects(-1),
 		nats.ReconnectWait(15*time.Millisecond),
-		nats.ReconnectHandler(func(_ *nats.Conn) {
-			ch <- true
+		nats.ConnectHandler(func(_ *nats.Conn) {
+			connectedCh <- true
 		}),
 		nats.NoCallbacksAfterClientClose())
 	if err != nil {
@@ -2752,27 +2935,17 @@ func TestRetryOnFailedConnectWithTLSError(t *testing.T) {
 	// Replace tls timeout to a reasonable value.
 	s.Shutdown()
 	opts.TLSTimeout = 2.0
-	s = RunServerWithOptions(opts)
+	s = RunServerWithOptions(&opts)
 	defer s.Shutdown()
 
 	select {
-	case <-ch:
+	case <-connectedCh:
 	case <-time.After(time.Second):
 		t.Fatal("Should have connected")
 	}
 }
 
 func TestConnStatusChangedEvents(t *testing.T) {
-	waitForStatus := func(t *testing.T, ch chan nats.Status, expected nats.Status) {
-		select {
-		case s := <-ch:
-			if s != expected {
-				t.Fatalf("Expected status: %s; got: %s", expected, s)
-			}
-		case <-time.After(5 * time.Second):
-			t.Fatalf("Timeout waiting for status %q", expected)
-		}
-	}
 	t.Run("default events", func(t *testing.T) {
 		s := RunDefaultServer()
 		nc, err := nats.Connect(s.ClientURL())
@@ -2795,15 +2968,15 @@ func TestConnStatusChangedEvents(t *testing.T) {
 		time.Sleep(50 * time.Millisecond)
 
 		s.Shutdown()
-		waitForStatus(t, newStatus, nats.RECONNECTING)
+		WaitOnChannel(t, newStatus, nats.RECONNECTING)
 
 		s = RunDefaultServer()
 		defer s.Shutdown()
 
-		waitForStatus(t, newStatus, nats.CONNECTED)
+		WaitOnChannel(t, newStatus, nats.CONNECTED)
 
 		nc.Close()
-		waitForStatus(t, newStatus, nats.CLOSED)
+		WaitOnChannel(t, newStatus, nats.CLOSED)
 
 		select {
 		case s := <-newStatus:
@@ -2836,7 +3009,7 @@ func TestConnStatusChangedEvents(t *testing.T) {
 		s = RunDefaultServer()
 		defer s.Shutdown()
 		nc.Close()
-		waitForStatus(t, newStatus, nats.CLOSED)
+		WaitOnChannel(t, newStatus, nats.CLOSED)
 
 		select {
 		case s := <-newStatus:
@@ -2863,3 +3036,112 @@ func TestConnStatusChangedEvents(t *testing.T) {
 		time.Sleep(100 * time.Millisecond)
 	})
 }
+
+func TestTLSHandshakeFirst(t *testing.T) {
+	s, opts := RunServerWithConfig("./configs/tls.conf")
+	defer s.Shutdown()
+
+	secureURL := fmt.Sprintf("tls://derek:porkchop@localhost:%d", opts.Port)
+	nc, err := nats.Connect(secureURL,
+		nats.RootCAs("./configs/certs/ca.pem"),
+		nats.TLSHandshakeFirst())
+	if err == nil || !strings.Contains(err.Error(), "TLS handshake") {
+		if err == nil {
+			nc.Close()
+		}
+		t.Fatalf("Expected error about not being a TLS handshake, got %v", err)
+	}
+
+	tc := &server.TLSConfigOpts{
+		CertFile: "./configs/certs/server.pem",
+		KeyFile:  "./configs/certs/key.pem",
+	}
+	tlsConf, err := server.GenTLSConfig(tc)
+	if err != nil {
+		t.Fatalf("Can't build TLCConfig: %v", err)
+	}
+	tlsConf.ServerName = "localhost"
+
+	// Start a mockup server that will do the TLS handshake first
+	// and then send the INFO protocol.
+	l, e := net.Listen("tcp", ":0")
+	if e != nil {
+		t.Fatal("Could not listen on an ephemeral port")
+	}
+	tl := l.(*net.TCPListener)
+	defer tl.Close()
+
+	addr := tl.Addr().(*net.TCPAddr)
+
+	errCh := make(chan error, 1)
+	doneCh := make(chan struct{})
+	wg := sync.WaitGroup{}
+	wg.Add(1)
+	go func() {
+		defer wg.Done()
+		conn, err := l.Accept()
+		if err != nil {
+			errCh <- fmt.Errorf("error accepting client connection: %v", err)
+			return
+		}
+		defer conn.Close()
+
+		// Do the TLS handshake now.
+		conn = tls.Server(conn, tlsConf)
+		tlsconn := conn.(*tls.Conn)
+		if err := tlsconn.Handshake(); err != nil {
+			errCh <- fmt.Errorf("Server error during handshake: %v", err)
+			return
+		}
+
+		// Send back the INFO
+		info := fmt.Sprintf("INFO {\"server_id\":\"foobar\",\"host\":\"localhost\",\"port\":%d,\"auth_required\":false,\"tls_required\":true,\"tls_available\":true,\"tls_verify\":true,\"max_payload\":1048576}\r\n", addr.Port)
+		tlsconn.Write([]byte(info))
+
+		// Read connect and ping commands sent from the client
+		line := make([]byte, 256)
+		_, err = tlsconn.Read(line)
+		if err != nil {
+			errCh <- fmt.Errorf("expected CONNECT and PING from client, got: %s", err)
+			return
+		}
+		tlsconn.Write([]byte("PONG\r\n"))
+
+		// Wait for the signal that client is ok
+		<-doneCh
+		// Server is done now.
+		errCh <- nil
+	}()
+
+	time.Sleep(100 * time.Millisecond)
+
+	secureURL = fmt.Sprintf("tls://derek:porkchop@localhost:%d", addr.Port)
+	nc, err = nats.Connect(secureURL,
+		nats.RootCAs("./configs/certs/ca.pem"),
+		nats.TLSHandshakeFirst())
+	if err != nil {
+		wg.Wait()
+		e := <-errCh
+		t.Fatalf("Unexpected error: %v (server error=%s)", err, e.Error())
+	}
+
+	state, err := nc.TLSConnectionState()
+	if err != nil {
+		t.Fatalf("Expected connection state: %v", err)
+	}
+	if !state.HandshakeComplete {
+		t.Fatalf("Expected valid connection state")
+	}
+	nc.Close()
+
+	close(doneCh)
+	wg.Wait()
+	select {
+	case e := <-errCh:
+		if e != nil {
+			t.Fatalf("Error from server: %v", err)
+		}
+	case <-time.After(2 * time.Second):
+		t.Fatal("Server did not exit")
+	}
+}
